package mindex.distance;

import messif.algorithms.AlgorithmMethodException;
import messif.objects.LocalAbstractObject;
import mindex.MIndexPPCalculator;
import mindex.MetricIndex;
import mindex.navigation.VoronoiInternalCell;

/**
 * This class calculates distance between a query object represented by query-pivot distances
 *  and a data object PPP using the "sum of query-pivot distance differences" normalized by
 *  the level (length of PPP).
 * 
 * @author David Novak, Masaryk University, Brno, Czech Republic, david.novak@fi.muni.cz
 */
public class QueryPivotDistDiffCalculator implements QueryPPPDistanceCalculator {

    /** Distance to all the pivots (can be set) */
    protected final float [] queryPivotDistances;
    
    /** PPP of the query */
    protected final short [] queryPP;

    protected volatile int distanceCalculator = 0;
    
    /** This field specifies the maximal level of PPP to be used (M-Index max level). */
    protected int maxLevel;
    
    @Override
    public float[] getQueryPivotDistances() {
        return queryPivotDistances;
    }

    @Override
    public short[] getQueryPPP() {
        return queryPP;
    }

    @Override
    public int getDistCalculationStat() {
        return distanceCalculator;
    }

    @Override
    public final int getMaxLevel() {
        return maxLevel;
    }

    @Override
    public final int getMaxLevel(int currentMax) {
        return (currentMax < maxLevel) ? currentMax : maxLevel;
    }

    @Override
    public final void setMaxLevel(int maxLevel) {
        this.maxLevel = maxLevel;
    }
        
    
    /**
     * Creates the calculator from given query object and the M-Index configuration
     * @param queryObject query object
     * @param mIndex M-Index configuration
     */
    public QueryPivotDistDiffCalculator(LocalAbstractObject queryObject, MetricIndex mIndex) throws AlgorithmMethodException {
        maxLevel = mIndex.getMaxLevel();
        queryPivotDistances = mIndex.initPPGetDistances(queryObject);
        queryPP = MIndexPPCalculator.sortDistances(queryPivotDistances, (short) queryPivotDistances.length);
        //queryPP = mIndex.readPPP(queryObject);
    }

    /**
     * For testing purposes 
     */
    protected QueryPivotDistDiffCalculator(short maxLevel, float [] queryPivotDistances, short [] queryPPP) {
        this.maxLevel = maxLevel;
        this.queryPivotDistances = queryPivotDistances;
        this.queryPP = queryPPP;
    }
    
    // ************** Query-PPP distance calculation ***************//

    @Override
    public float getQueryDistance(short[] ppp) {        
//        distanceCalculator 
        float sum = 0f;
        for (int i = 0; i < getMaxLevel(ppp.length); i++) {
            sum += Math.max(queryPivotDistances[ppp[i]] - queryPivotDistances[queryPP[i]], 0f);
        }
        return sum;
        //return sum / (float) ppp.length;
    }

    @Override
    public float getQueryDistance(final VoronoiInternalCell intCell) {
        return getQueryDistance(intCell.getPppForReading());
    }    
}
